/**
 * Copyright 2018, OpenCensus Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/// <reference types="node" />
import { BasePlugin, Func, SpanKind, TagMap } from '@opencensus/core';
import { ClientRequest, IncomingMessage, request, RequestOptions } from 'http';
import * as url from 'url';
import { HttpPluginConfig, IgnoreMatcher } from './types';
export declare type HttpGetCallback = (res: IncomingMessage) => void;
export declare type RequestFunction = typeof request;
/** Http instrumentation plugin for Opencensus */
export declare class HttpPlugin extends BasePlugin {
    /**
     * Attributes Names according to Opencensus HTTP Specs
     * https://github.com/census-instrumentation/opencensus-specs/blob/master/trace/HTTP.md
     */
    static ATTRIBUTE_HTTP_HOST: string;
    static ATTRIBUTE_HTTP_METHOD: string;
    static ATTRIBUTE_HTTP_PATH: string;
    static ATTRIBUTE_HTTP_ROUTE: string;
    static ATTRIBUTE_HTTP_USER_AGENT: string;
    static ATTRIBUTE_HTTP_STATUS_CODE: string;
    static ATTRIBUTE_HTTP_ERROR_NAME: string;
    static ATTRIBUTE_HTTP_ERROR_MESSAGE: string;
    options: HttpPluginConfig;
    /** Constructs a new HttpPlugin instance. */
    constructor(moduleName: string);
    /** Patches HTTP incoming and outcoming request functions. */
    protected applyPatch(): any;
    /** Unpatches all HTTP patched function. */
    protected applyUnpatch(): void;
    /**
     * Check whether the given request is ignored by configuration
     * @param url URL of request
     * @param request Request to inspect
     * @param list List of ignore patterns
     */
    protected isIgnored<T>(url: string, request: T, list?: Array<IgnoreMatcher<T>>): boolean;
    /**
     * Check whether the given request match pattern
     * @param url URL of request
     * @param request Request to inspect
     * @param pattern Match pattern
     */
    protected isSatisfyPattern<T>(url: string, request: T, pattern: IgnoreMatcher<T>): boolean;
    /**
     * Creates spans for incoming requests, restoring spans' context if applied.
     */
    protected getPatchIncomingRequestFunction(): (original: (event: string) => boolean) => (event: string, ...args: any[]) => boolean;
    /**
     * Creates spans for outgoing requests, sending spans' context for distributed
     * tracing.
     */
    protected getPatchOutgoingRequestFunction(): (original: Func<ClientRequest>) => Func<ClientRequest>;
    /**
     * Injects span's context to header for distributed tracing and finshes the
     * span when the response is finished.
     * @param original The original patched function.
     * @param options The arguments to the original function.
     */
    private getMakeRequestTraceFunction;
    /**
     * Parse OpenCensus Status from HTTP response status code.
     * @param statusCode The HTTP response status code.
     */
    static parseResponseStatus(statusCode: number): number;
    /** Method to record stats for client and server. */
    static recordStats(kind: SpanKind, tags: TagMap, ms: number): void;
    /**
     * Returns whether the Expect header is on the given options object.
     * @param options Options for http.request.
     */
    hasExpectHeader(options: RequestOptions | url.URL): boolean;
}
declare const plugin: HttpPlugin;
export { plugin };
