"use strict";
/**
 * Copyright 2018, OpenCensus Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("@opencensus/core");
const http_1 = require("http");
const semver = require("semver");
const shimmer = require("shimmer");
const url = require("url");
const stats = require("./http-stats");
function isOpenCensusRequest(options) {
    return (options &&
        options.headers &&
        !!options.headers['x-opencensus-outgoing-request']);
}
const UNLIMITED_PROPAGATION_MD = {
    tagTtl: core_1.TagTtl.UNLIMITED_PROPAGATION,
};
const TAG_VALUE_MAX_LENGTH = 255;
/** Http instrumentation plugin for Opencensus */
class HttpPlugin extends core_1.BasePlugin {
    /** Constructs a new HttpPlugin instance. */
    constructor(moduleName) {
        super(moduleName);
    }
    /** Patches HTTP incoming and outcoming request functions. */
    applyPatch() {
        this.logger.debug('applying patch to %s@%s', this.moduleName, this.version);
        shimmer.wrap(this.moduleExports, 'request', this.getPatchOutgoingRequestFunction());
        // In Node 8, http.get calls a private request method, therefore we patch it
        // here too.
        if (semver.satisfies(this.version, '>=8.0.0')) {
            shimmer.wrap(this.moduleExports, 'get', () => {
                // Re-implement http.get. This needs to be done (instead of using
                // getPatchOutgoingRequestFunction to patch it) because we need to
                // set the trace context header before the returned ClientRequest is
                // ended. The Node.js docs state that the only differences between
                // request and get are that (1) get defaults to the HTTP GET method and
                // (2) the returned request object is ended immediately. The former is
                // already true (at least in supported Node versions up to v9), so we
                // simply follow the latter. Ref:
                // https://nodejs.org/dist/latest/docs/api/http.html#http_http_get_options_callback
                // https://github.com/googleapis/cloud-trace-nodejs/blob/master/src/plugins/plugin-http.ts#L198
                return function getTrace(options, callback) {
                    const req = http_1.request(options, callback);
                    req.end();
                    return req;
                };
            });
        }
        if (this.moduleExports &&
            this.moduleExports.Server &&
            this.moduleExports.Server.prototype) {
            shimmer.wrap(this.moduleExports.Server.prototype, 'emit', this.getPatchIncomingRequestFunction());
        }
        else {
            this.logger.error('Could not apply patch to %s.emit. Interface is not as expected.', this.moduleName);
        }
        return this.moduleExports;
    }
    /** Unpatches all HTTP patched function. */
    applyUnpatch() {
        shimmer.unwrap(this.moduleExports, 'request');
        if (semver.satisfies(this.version, '>=8.0.0')) {
            shimmer.unwrap(this.moduleExports, 'get');
        }
        if (this.moduleExports &&
            this.moduleExports.Server &&
            this.moduleExports.Server.prototype) {
            shimmer.unwrap(this.moduleExports.Server.prototype, 'emit');
        }
    }
    /**
     * Check whether the given request is ignored by configuration
     * @param url URL of request
     * @param request Request to inspect
     * @param list List of ignore patterns
     */
    isIgnored(url, request, list) {
        if (!list) {
            // No ignored urls - trace everything
            return false;
        }
        for (const pattern of list) {
            if (this.isSatisfyPattern(url, request, pattern)) {
                return true;
            }
        }
        return false;
    }
    /**
     * Check whether the given request match pattern
     * @param url URL of request
     * @param request Request to inspect
     * @param pattern Match pattern
     */
    isSatisfyPattern(url, request, pattern) {
        if (typeof pattern === 'string') {
            return pattern === url;
        }
        else if (pattern instanceof RegExp) {
            return pattern.test(url);
        }
        else if (typeof pattern === 'function') {
            return pattern(url, request);
        }
        else {
            throw new TypeError('Pattern is in unsupported datatype');
        }
    }
    /**
     * Creates spans for incoming requests, restoring spans' context if applied.
     */
    getPatchIncomingRequestFunction() {
        return (original) => {
            const plugin = this;
            // This function's signature is that of an event listener, which can have
            // any number of variable-type arguments.
            // tslint:disable-next-line:no-any
            return function incomingRequest(event, ...args) {
                // Only traces request events
                if (event !== 'request') {
                    return original.apply(this, arguments);
                }
                const startTime = Date.now();
                const request = args[0];
                const response = args[1];
                const path = request.url ? url.parse(request.url).pathname || '' : '';
                const method = request.method || 'GET';
                plugin.logger.debug('%s plugin incomingRequest', plugin.moduleName);
                if (plugin.isIgnored(path, request, plugin.options.ignoreIncomingPaths)) {
                    return original.apply(this, arguments);
                }
                const propagation = plugin.tracer.propagation;
                const headers = request.headers;
                const getter = {
                    getHeader(name) {
                        return headers[name];
                    },
                };
                const traceOptions = {
                    name: path,
                    kind: core_1.SpanKind.SERVER,
                };
                if (propagation) {
                    const spanContext = propagation.extract(getter);
                    if (spanContext) {
                        traceOptions.spanContext = spanContext;
                    }
                }
                return plugin.tracer.startRootSpan(traceOptions, rootSpan => {
                    if (!rootSpan)
                        return original.apply(this, arguments);
                    plugin.tracer.wrapEmitter(request);
                    plugin.tracer.wrapEmitter(response);
                    // Wraps end (inspired by:
                    // https://github.com/GoogleCloudPlatform/cloud-trace-nodejs/blob/master/src/plugins/plugin-connect.ts#L75)
                    const originalEnd = response.end;
                    response.end = function () {
                        response.end = originalEnd;
                        const returned = response.end.apply(this, arguments);
                        const requestUrl = request.url ? url.parse(request.url) : null;
                        const host = headers.host || 'localhost';
                        const userAgent = (headers['user-agent'] ||
                            headers['User-Agent']);
                        const tags = new core_1.TagMap();
                        rootSpan.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_HOST, host.replace(/^(.*)(\:[0-9]{1,5})/, '$1'));
                        rootSpan.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_METHOD, method);
                        if (requestUrl) {
                            rootSpan.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_PATH, requestUrl.pathname || '');
                            rootSpan.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_ROUTE, requestUrl.path || '');
                            tags.set(stats.HTTP_SERVER_ROUTE, {
                                value: (requestUrl.path || '').substring(0, TAG_VALUE_MAX_LENGTH),
                            }, UNLIMITED_PROPAGATION_MD);
                        }
                        if (userAgent) {
                            rootSpan.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_USER_AGENT, userAgent);
                        }
                        rootSpan.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_STATUS_CODE, response.statusCode.toString());
                        rootSpan.setStatus(HttpPlugin.parseResponseStatus(response.statusCode));
                        rootSpan.addMessageEvent(core_1.MessageEventType.RECEIVED, 1);
                        if (plugin.options.applyCustomAttributesOnSpan) {
                            plugin.options.applyCustomAttributesOnSpan(rootSpan, request, response);
                        }
                        tags.set(stats.HTTP_SERVER_METHOD, { value: method }, UNLIMITED_PROPAGATION_MD);
                        tags.set(stats.HTTP_SERVER_STATUS, { value: response.statusCode.toString() }, UNLIMITED_PROPAGATION_MD);
                        HttpPlugin.recordStats(rootSpan.kind, tags, Date.now() - startTime);
                        rootSpan.end();
                        return returned;
                    };
                    return original.apply(this, arguments);
                });
            };
        };
    }
    /**
     * Creates spans for outgoing requests, sending spans' context for distributed
     * tracing.
     */
    getPatchOutgoingRequestFunction() {
        return (original) => {
            const plugin = this;
            return function outgoingRequest(options, callback) {
                if (!options) {
                    return original.apply(this, [options, callback]);
                }
                // Makes sure the url is an url object
                let pathname;
                let method;
                let origin = '';
                if (typeof options === 'string') {
                    const parsedUrl = url.parse(options);
                    options = parsedUrl;
                    pathname = parsedUrl.pathname || '';
                    origin = `${parsedUrl.protocol || 'http:'}//${parsedUrl.host}`;
                }
                else {
                    // Do not trace ourselves
                    if (isOpenCensusRequest(options)) {
                        plugin.logger.debug('header with "x-opencensus-outgoing-request" - do not trace');
                        return original.apply(this, [options, callback]);
                    }
                    try {
                        pathname = options.pathname;
                        if (!pathname) {
                            pathname = options.path ? url.parse(options.path).pathname : '';
                        }
                        method = options.method || 'GET';
                        origin = `${options.protocol || 'http:'}//${options.host}`;
                    }
                    catch (ignore) { }
                }
                const request = original.apply(this, [
                    options,
                    callback,
                ]);
                if (plugin.isIgnored(origin + pathname, request, plugin.options.ignoreOutgoingUrls)) {
                    return request;
                }
                plugin.tracer.wrapEmitter(request);
                plugin.logger.debug('%s plugin outgoingRequest', plugin.moduleName);
                const traceOptions = {
                    name: `${method || 'GET'} ${pathname}`,
                    kind: core_1.SpanKind.CLIENT,
                };
                // Checks if this outgoing request is part of an operation by checking
                // if there is a current root span, if so, we create a child span. In
                // case there is no root span, this means that the outgoing request is
                // the first operation, therefore we create a root span.
                if (!plugin.tracer.currentRootSpan) {
                    plugin.logger.debug('outgoingRequest starting a root span');
                    return plugin.tracer.startRootSpan(traceOptions, plugin.getMakeRequestTraceFunction(request, options, plugin));
                }
                else {
                    plugin.logger.debug('outgoingRequest starting a child span');
                    const span = plugin.tracer.startChildSpan({
                        name: traceOptions.name,
                        kind: traceOptions.kind,
                    });
                    return plugin.getMakeRequestTraceFunction(request, options, plugin)(span);
                }
            };
        };
    }
    /**
     * Injects span's context to header for distributed tracing and finshes the
     * span when the response is finished.
     * @param original The original patched function.
     * @param options The arguments to the original function.
     */
    getMakeRequestTraceFunction(request, options, plugin) {
        return (span) => {
            const startTime = Date.now();
            plugin.logger.debug('makeRequestTrace');
            if (!span) {
                plugin.logger.debug('makeRequestTrace span is null');
                return request;
            }
            const setter = {
                setHeader(name, value) {
                    // If outgoing request headers contain the "Expect" header, the
                    // returned ClientRequest will throw an error if any new headers are
                    // added. We need to set the header directly in the headers object
                    // which has been cloned earlier.
                    if (plugin.hasExpectHeader(options) && options.headers) {
                        options.headers[name] = value;
                    }
                    else {
                        request.setHeader(name, value);
                    }
                },
            };
            const propagation = plugin.tracer.propagation;
            if (propagation) {
                // If outgoing request headers contain the "Expect" header, the returned
                // ClientRequest will throw an error if any new headers are added.
                // So we need to clone the options object to be able to inject new
                // header.
                if (plugin.hasExpectHeader(options)) {
                    options = Object.assign({}, options);
                    options.headers = Object.assign({}, options.headers);
                }
                propagation.inject(setter, span.spanContext);
            }
            request.on('response', (response) => {
                plugin.tracer.wrapEmitter(response);
                plugin.logger.debug('outgoingRequest on response()');
                response.on('end', () => {
                    plugin.logger.debug('outgoingRequest on end()');
                    const method = response.method ? response.method : 'GET';
                    const headers = options.headers;
                    const userAgent = headers
                        ? headers['user-agent'] || headers['User-Agent']
                        : null;
                    const tags = new core_1.TagMap();
                    tags.set(stats.HTTP_CLIENT_METHOD, { value: method });
                    const host = options.hostname || options.host || 'localhost';
                    span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_HOST, host);
                    span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_METHOD, method);
                    if (options.path) {
                        span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_PATH, options.path);
                        span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_ROUTE, options.path);
                    }
                    if (userAgent) {
                        span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_USER_AGENT, userAgent.toString());
                    }
                    if (response.statusCode) {
                        span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_STATUS_CODE, response.statusCode.toString());
                        span.setStatus(HttpPlugin.parseResponseStatus(response.statusCode));
                        tags.set(stats.HTTP_CLIENT_STATUS, {
                            value: response.statusCode.toString(),
                        });
                    }
                    span.addMessageEvent(core_1.MessageEventType.SENT, 1);
                    if (plugin.options.applyCustomAttributesOnSpan) {
                        plugin.options.applyCustomAttributesOnSpan(span, request, response);
                    }
                    HttpPlugin.recordStats(span.kind, tags, Date.now() - startTime);
                    span.end();
                });
                response.on('error', error => {
                    span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_ERROR_NAME, error.name);
                    span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_ERROR_MESSAGE, error.message);
                    span.setStatus(core_1.CanonicalCode.UNKNOWN, error.message);
                    span.end();
                });
            });
            request.on('error', error => {
                span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_ERROR_NAME, error.name);
                span.addAttribute(HttpPlugin.ATTRIBUTE_HTTP_ERROR_MESSAGE, error.message);
                span.setStatus(core_1.CanonicalCode.UNKNOWN, error.message);
                span.end();
            });
            plugin.logger.debug('makeRequestTrace return request');
            return request;
        };
    }
    /**
     * Parse OpenCensus Status from HTTP response status code.
     * @param statusCode The HTTP response status code.
     */
    static parseResponseStatus(statusCode) {
        if (statusCode < 200 || statusCode > 504) {
            return core_1.CanonicalCode.UNKNOWN;
        }
        else if (statusCode >= 200 && statusCode < 400) {
            return core_1.CanonicalCode.OK;
        }
        else {
            switch (statusCode) {
                case 400:
                    return core_1.CanonicalCode.INVALID_ARGUMENT;
                case 504:
                    return core_1.CanonicalCode.DEADLINE_EXCEEDED;
                case 404:
                    return core_1.CanonicalCode.NOT_FOUND;
                case 403:
                    return core_1.CanonicalCode.PERMISSION_DENIED;
                case 401:
                    return core_1.CanonicalCode.UNAUTHENTICATED;
                case 429:
                    return core_1.CanonicalCode.RESOURCE_EXHAUSTED;
                case 501:
                    return core_1.CanonicalCode.UNIMPLEMENTED;
                case 503:
                    return core_1.CanonicalCode.UNAVAILABLE;
                default:
                    return core_1.CanonicalCode.UNKNOWN;
            }
        }
    }
    /** Method to record stats for client and server. */
    static recordStats(kind, tags, ms) {
        if (!plugin.stats)
            return;
        try {
            const measureList = [];
            switch (kind) {
                case core_1.SpanKind.CLIENT:
                    measureList.push({
                        measure: stats.HTTP_CLIENT_ROUNDTRIP_LATENCY,
                        value: ms,
                    });
                    break;
                case core_1.SpanKind.SERVER:
                    measureList.push({ measure: stats.HTTP_SERVER_LATENCY, value: ms });
                    break;
                default:
                    break;
            }
            plugin.stats.record(measureList, tags);
        }
        catch (ignore) { }
    }
    /**
     * Returns whether the Expect header is on the given options object.
     * @param options Options for http.request.
     */
    hasExpectHeader(options) {
        return !!(options.headers &&
            options.headers.Expect);
    }
}
/**
 * Attributes Names according to Opencensus HTTP Specs
 * https://github.com/census-instrumentation/opencensus-specs/blob/master/trace/HTTP.md
 */
HttpPlugin.ATTRIBUTE_HTTP_HOST = 'http.host';
HttpPlugin.ATTRIBUTE_HTTP_METHOD = 'http.method';
HttpPlugin.ATTRIBUTE_HTTP_PATH = 'http.path';
HttpPlugin.ATTRIBUTE_HTTP_ROUTE = 'http.route';
HttpPlugin.ATTRIBUTE_HTTP_USER_AGENT = 'http.user_agent';
HttpPlugin.ATTRIBUTE_HTTP_STATUS_CODE = 'http.status_code';
// NOT ON OFFICIAL SPEC
HttpPlugin.ATTRIBUTE_HTTP_ERROR_NAME = 'http.error_name';
HttpPlugin.ATTRIBUTE_HTTP_ERROR_MESSAGE = 'http.error_message';
exports.HttpPlugin = HttpPlugin;
const plugin = new HttpPlugin('http');
exports.plugin = plugin;
//# sourceMappingURL=http.js.map